@@ -21,10 +21,9 @@
 
 #include "nm-default.h"
 
-#include <string.h>
+#include "nm-active-connection.h"
 
 #include "nm-dbus-interface.h"
-#include "nm-active-connection.h"
 #include "nm-object-private.h"
 #include "nm-core-internal.h"
 #include "nm-device.h"
@@ -580,7 +579,6 @@ init_dbus (NMObject *object)
 	                                property_info);
 }
 
-
 static void
 nm_active_connection_class_init (NMActiveConnectionClass *ap_class)
 {
@@ -661,11 +659,9 @@ nm_active_connection_class_init (NMActiveConnectionClass *ap_class)
 		                      G_PARAM_STATIC_STRINGS));
 
 	/**
-	 * NMActiveConnection:devices:
+	 * NMActiveConnection:devices: (type GPtrArray(NMDevice))
 	 *
 	 * The devices of the active connection.
-	 *
-	 * Element-type: NMDevice
 	 **/
 	g_object_class_install_property
 		(object_class, PROP_DEVICES,
@@ -798,7 +794,12 @@ nm_active_connection_class_init (NMActiveConnectionClass *ap_class)
 		                      G_PARAM_READABLE |
 		                      G_PARAM_STATIC_STRINGS));
 
-	/* signals */
+	/**
+	 * NMActiveConnection::state-changed:
+	 * @active_connection: the source #NMActiveConnection
+	 * @state: the new state number (#NMActiveConnectionState)
+	 * @reason: the state change reason (#NMActiveConnectionStateReason)
+	 */
 	signals[STATE_CHANGED] =
 		g_signal_new ("state-changed",
 		              G_OBJECT_CLASS_TYPE (object_class),
